from functools import reduce
from math import ceil
from typing import Any, Dict, Union

from terra_sdk.core import Coin

from .api._base import BaseAsyncAPI, sync_bind


def index_by_pub_key(m: Dict[str, Any], o: Any):
    m[o["pub_key"]["value"]] = o
    return m


class AsyncLCDUtils(BaseAsyncAPI):
    async def calculate_tax(self, coin: Union[Coin, str, dict]) -> Coin:
        """Calculates the tax that would be applied for the Coin if sent.

        Args:
            coin (Union[Coin, str, dict]): coin to be sent

        Returns:
            Coin: tax to be paid
        """
        coin = Coin.parse(coin)
        rate = await BaseAsyncAPI._try_await(self._c.treasury.tax_rate())
        cap = await BaseAsyncAPI._try_await(self._c.treasury.tax_cap(coin.denom))
        return Coin(coin.denom, min(ceil(coin.amount * rate), cap.amount))

    async def validators_with_voting_power(self) -> Dict[str, dict]:
        """Gets current validators and merges their voting power from the validator set query.

        Returns:
            Dict[str, dict]: validators with voting power
        """
        validator_set_response = await BaseAsyncAPI._try_await(
            self._c.tendermint.validator_set()
        )
        validators = await BaseAsyncAPI._try_await(self._c.staking.validators())
        validator_set: Dict[str, Any] = reduce(
            index_by_pub_key, validator_set_response["validators"], {}
        )
        res = {}
        for v in validators:
            delegate_info = validator_set.get(v.consensus_pubkey["value"])
            if delegate_info is None:
                continue
            res[v.operator_address] = {
                "validator_info": v,
                "voting_power": int(delegate_info["voting_power"]),
                "proposer_priority": int(delegate_info["proposer_priority"]),
            }
        return res


class LCDUtils(AsyncLCDUtils):
    @sync_bind(AsyncLCDUtils.calculate_tax)
    def calculate_tax(self, coin: Union[Coin, str, dict]) -> Coin:
        pass

    calculate_tax.__doc__ = AsyncLCDUtils.calculate_tax.__doc__

    @sync_bind(AsyncLCDUtils.validators_with_voting_power)
    async def validators_with_voting_power(self) -> Dict[str, dict]:
        pass

    validators_with_voting_power.__doc__ = AsyncLCDUtils.calculate_tax.__doc__



from .client import Height, IdentifiedClientState, ConsensusStateWithHeight, ClientConsensusStates, Params
from .commitment import MerkleRoot, MerklePrefix
from .channel import Counterparty, Channel, Order, State, Packet

__all__ = [
    "Height",
    "IdentifiedClientState",
    "ConsensusStateWithHeight",
    "ClientConsensusStates",
    "Params",
    "MerkleRoot",
    "MerklePrefix",
    "Counterparty",
    "Channel",
    "Order",
    "State",
    "Packet"
]



from typing import Optional

from terra_sdk.core import Coin, Coins, Dec, Numeric
from terra_sdk.core.treasury import PolicyConstraints

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncTreasuryAPI", "TreasuryAPI"]


class AsyncTreasuryAPI(BaseAsyncAPI):
    async def tax_cap(self, denom: str) -> Coin:
        """Fetches the tax cap for a denom.

        Args:
            denom (str): denom

        Returns:
            Coin: tax cap
        """
        res = await self._c._get(f"/terra/treasury/v1beta1/tax_caps/{denom}")
        return Coin(denom, res.get("tax_cap"))

    async def tax_rate(self, height: Optional[int] = None) -> Dec:
        """Fetches the current tax rate.

        Returns:
            Dec: tax rate
        """
        params = None
        if height is not None:
            params = {"height": height}
        res = await self._c._get("/terra/treasury/v1beta1/tax_rate", params)
        return Dec(res.get("tax_rate"))

    async def reward_weight(self) -> Dec:
        """Fetches the current reward rate.

        Returns:
            Dec: reward weight
        """
        res = await self._c._get("//terra/treasury/v1beta1/reward_weight")
        return Dec(res.get("reward_weight"))

    async def tax_proceeds(self) -> Coins:
        """Fetches the current tax proceeds.

        Returns:
            Coins: tax proceeds
        """
        res = await self._c._get("/terra/treasury/v1beta1/tax_proceeds")
        return Coins.from_data(res.get("tax_proceeds"))

    async def seigniorage_proceeds(self) -> Coin:
        """Fetches the current seigniorage proceeds.

        Returns:
            Coin: seigniorage proceeds
        """
        res = await self._c._get("/terra/treasury/v1beta1/seigniorage_proceeds")
        return Coin("uluna", res.get("seigniorage_proceeds"))

    async def parameters(self) -> Coin:
        """Fetches the Treasury module parameters.

        Returns:
            Coin: Treasury module parameters.
        """
        res = await self._c._get("/terra/treasury/v1beta1/params")
        params = res.get("params")
        return {
            "tax_policy": PolicyConstraints.from_data(params["tax_policy"]),
            "reward_policy": PolicyConstraints.from_data(params["reward_policy"]),
            "mining_increment": Dec(params["mining_increment"]),
            "seigniorage_burden_target": Dec(params["seigniorage_burden_target"]),
            "window_long": Numeric.parse(params["window_long"]),
            "window_short": Numeric.parse(params["window_short"]),
            "window_probation": Numeric.parse(params["window_probation"]),
        }


class TreasuryAPI(AsyncTreasuryAPI):
    @sync_bind(AsyncTreasuryAPI.tax_cap)
    def tax_cap(self, denom: str) -> Coin:
        pass

    tax_cap.__doc__ = AsyncTreasuryAPI.tax_cap.__doc__

    @sync_bind(AsyncTreasuryAPI.tax_rate)
    def tax_rate(self, height: Optional[int] = None) -> Dec:
        pass

    tax_rate.__doc__ = AsyncTreasuryAPI.tax_rate.__doc__

    @sync_bind(AsyncTreasuryAPI.reward_weight)
    def reward_weight(self) -> Dec:
        pass

    reward_weight.__doc__ = AsyncTreasuryAPI.reward_weight.__doc__

    @sync_bind(AsyncTreasuryAPI.tax_proceeds)
    def tax_proceeds(self) -> Coins:
        pass

    tax_proceeds.__doc__ = AsyncTreasuryAPI.tax_proceeds.__doc__

    @sync_bind(AsyncTreasuryAPI.seigniorage_proceeds)
    def seigniorage_proceeds(self) -> Coin:
        pass

    seigniorage_proceeds.__doc__ = AsyncTreasuryAPI.seigniorage_proceeds.__doc__

    @sync_bind(AsyncTreasuryAPI.parameters)
    def parameters(self) -> Coin:
        pass

    parameters.__doc__ = AsyncTreasuryAPI.parameters.__doc__



from abc import ABC, abstractmethod

from terra_sdk.core.public_key import PublicKey
from terra_sdk.util.json import JSONSerializable

from .base_account import BaseAccount
from .lazy_graded_vesting_account import LazyGradedVestingAccount


class Account(JSONSerializable):

    @abstractmethod
    def get_account_number(self) -> int:
        pass

    @abstractmethod
    def get_sequence(self) -> int:
        pass

    @abstractmethod
    def get_public_key(self) -> PublicKey:
        pass

    @classmethod
    def from_data(cls, data: dict):  # -> Account:
        if data["@type"] == BaseAccount.type_url:
            return BaseAccount.from_data(data)
        else:
            return LazyGradedVestingAccount.from_data(data)



"Wasm module messages."

from __future__ import annotations

import base64
import copy
import json
from typing import Optional

import attr
from terra_proto.terra.wasm.v1beta1 import (
    MsgClearContractAdmin as MsgClearContractAdmin_pb,
)
from terra_proto.terra.wasm.v1beta1 import MsgExecuteContract as MsgExecuteContract_pb
from terra_proto.terra.wasm.v1beta1 import (
    MsgInstantiateContract as MsgInstantiateContract_pb,
)
from terra_proto.terra.wasm.v1beta1 import MsgMigrateCode as MsgMigrateCode_pb
from terra_proto.terra.wasm.v1beta1 import MsgMigrateContract as MsgMigrateContract_pb
from terra_proto.terra.wasm.v1beta1 import MsgStoreCode as MsgStoreCode_pb
from terra_proto.terra.wasm.v1beta1 import (
    MsgUpdateContractAdmin as MsgUpdateContractAdmin_pb,
)

from terra_sdk.core import AccAddress, Coins
from terra_sdk.core.msg import Msg
from terra_sdk.util.json import dict_to_data
from terra_sdk.util.remove_none import remove_none

__all__ = [
    "MsgStoreCode",
    "MsgMigrateCode",
    "MsgInstantiateContract",
    "MsgExecuteContract",
    "MsgMigrateContract",
    "MsgUpdateContractAdmin",
    "MsgClearContractAdmin",
]


@attr.s
class MsgStoreCode(Msg):
    """Upload a new smart contract WASM binary to the blockchain.

    Args:
        sender: address of sender
        wasm_byte_code: base64-encoded string containing bytecode
    """

    type_amino = "wasm/MsgStoreCode"
    """"""
    type_url = "/terra.wasm.v1beta1.MsgStoreCode"
    """"""

    sender: AccAddress = attr.ib()
    wasm_byte_code: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "wasm_byte_code": self.wasm_byte_code
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgStoreCode:
        return cls(sender=data["sender"], wasm_byte_code=data["wasm_byte_code"])

    def to_proto(self) -> MsgStoreCode_pb:
        return MsgStoreCode_pb(
            sender=self.sender, wasm_byte_code=base64.b64decode(self.wasm_byte_code)
        )

    @classmethod
    def from_proto(cls, proto: MsgStoreCode_pb) -> MsgStoreCode:
        return cls(
            sender=proto.sender,
            wasm_byte_code=proto.wasm_byte_code
        )


@attr.s
class MsgMigrateCode(Msg):
    """Upload a new smart contract WASM binary to the blockchain, replacing an existing code ID.
    Can only be called once by creator of the contract, and is used for migrating from Col-4 to Col-5.

    Args:
        sender: address of sender
        code_id: reference to the code on the blockchain
        wasm_byte_code: base64-encoded string containing bytecode
    """

    type_amino = "wasm/MsgMigrateCode"
    """"""
    type_url = "/terra.wasm.v1beta1.MsgMigrateCode"
    """"""

    sender: AccAddress = attr.ib()
    code_id: int = attr.ib(converter=int)
    wasm_byte_code: str = attr.ib(converter=str)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "code_id": str(self.code_id),
                "wasm_byte_code": self.wasm_byte_code
            }
        }

    def to_data(self) -> dict:
        d = copy.deepcopy(self.__dict__)
        d["sender"] = str(d["sender"])
        d["code_id"] = str(d["code_id"])
        d["wasm_byte_code"] = str(d["wasm_byte_code"])
        return {"type": self.type, "value": dict_to_data(d)}

    @classmethod
    def from_data(cls, data: dict) -> MsgMigrateCode:
        return cls(
            sender=data["sender"],
            code_id=data["code_id"],
            wasm_byte_code=data["wasm_byte_code"],
        )

    def to_proto(self) -> MsgMigrateCode_pb:
        return MsgMigrateCode_pb(
            sender=self.sender, code_id=self.code_id, wasm_byte_code=self.wasm_byte_code
        )

    @classmethod
    def from_proto(cls, proto: MsgMigrateCode_pb) -> MsgMigrateCode:
        return cls(
            sender=proto.sender,
            code_id=proto.code_id,
            wasm_byte_code=proto.wasm_byte_code,
        )

@attr.s
class MsgInstantiateContract(Msg):
    """Creates a new instance of a smart contract from existing code on the blockchain.

    Args:
        sender: address of sender
        admin: address of contract admin
        code_id (int): code ID to use for instantiation
        init_msg (dict): InitMsg to initialize contract
        init_coins (Coins): initial amount of coins to be sent to contract
    """

    type_amino = "wasm/MsgInstantiateContract"
    """"""
    type_url = "/terra.wasm.v1beta1.MsgInstantiateContract"
    """"""

    sender: AccAddress = attr.ib()
    admin: Optional[AccAddress] = attr.ib()
    code_id: int = attr.ib(converter=int)
    init_msg: dict = attr.ib()
    init_coins: Coins = attr.ib(converter=Coins, factory=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "admin": self.admin,
                "code_id": str(self.code_id),
                "init_msg": self.init_msg,
                "init_coins": self.init_coins.to_amino()
            }
        }

    def to_data(self) -> dict:
        d = copy.deepcopy(self.__dict__)
        d["code_id"] = str(d["code_id"])
        return {"type": self.type, "value": dict_to_data(d)}

    @classmethod
    def from_data(cls, data: dict) -> MsgInstantiateContract:
        return cls(
            sender=data.get("sender"),
            admin=data.get("admin"),
            code_id=data["code_id"],
            init_msg=remove_none(data["init_msg"]),
            init_coins=Coins.from_data(data["init_coins"]),
        )

    def to_proto(self) -> MsgInstantiateContract_pb:
        return MsgInstantiateContract_pb(
            sender=self.sender,
            admin=self.admin,
            code_id=self.code_id,
            init_msg=bytes(json.dumps(self.init_msg), "utf-8"),
            init_coins=self.init_coins.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgInstantiateContract_pb) -> MsgInstantiateContract:
        return cls(
            sender=proto.sender,
            admin=proto.admin,
            code_id=proto.code_id,
            init_msg=remove_none(proto.init_msg),
            init_coins=Coins.from_proto(proto.init_coins),
        )



@attr.s
class MsgExecuteContract(Msg):
    """Execute a state-mutating function on a smart contract.

    Args:
        sender: address of sender
        contract: address of contract to execute function on
        execute_msg (dict): ExecuteMsg to pass
        coins: coins to be sent, if needed by contract to execute.
            Defaults to empty ``Coins()``
    """

    type_amino = "wasm/MsgExecuteContract"
    """"""
    type_url = "/terra.wasm.v1beta1.MsgExecuteContract"
    """"""

    sender: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()
    execute_msg: dict = attr.ib()
    coins: Coins = attr.ib(converter=Coins, factory=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "sender": self.sender,
                "contract": self.contract,
                "execute_msg": remove_none(self.execute_msg),
                "coins": self.coins.to_amino()
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgExecuteContract:
        return cls(
            sender=data["sender"],
            contract=data["contract"],
            execute_msg=remove_none(data.get("execute_msg")),
            coins=Coins.from_data(data["coins"]),
        )

    def to_proto(self) -> MsgExecuteContract_pb:
        return MsgExecuteContract_pb(
            sender=self.sender,
            contract=self.contract,
            execute_msg=bytes(json.dumps(self.execute_msg), "utf-8"),
            coins=self.coins.to_proto(),
        )

    @classmethod
    def from_proto(cls, proto: MsgExecuteContract_pb) -> MsgExecuteContract:
        return cls(
            sender=proto.sender,
            contract=proto.contract,
            execute_msg=remove_none(proto.execute_msg),
            coins=Coins.from_proto(proto.coins),
        )


@attr.s
class MsgMigrateContract(Msg):
    """Migrate the contract to a different code ID.

    Args:
        admin: address of contract admin
        contract: address of contract to migrate
        new_code_id (int): new code ID to migrate to
        migrate_msg (dict): MigrateMsg to execute
    """

    type_amino = "wasm/MsgMigrateContract"
    """"""
    type_url = "/terra.wasm.v1beta1.MsgMigrateContract"
    """"""

    admin: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()
    new_code_id: int = attr.ib(converter=int)
    migrate_msg: dict = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "admin": self.admin,
                "contract": self.contract,
                "new_code_id": str(self.new_code_id),
                "migrate_msg": remove_none(self.migrate_msg)
            }
        }

    def to_data(self) -> dict:
        d = copy.deepcopy(self.__dict__)
        return {"type": self.type, "value": dict_to_data(d)}

    @classmethod
    def from_data(cls, data: dict) -> MsgMigrateContract:
        return cls(
            admin=data["admin"],
            contract=data["contract"],
            new_code_id=data["new_code_id"],
            migrate_msg=data["migrate_msg"],
        )

    def to_proto(self) -> MsgMigrateContract_pb:
        return MsgMigrateContract_pb(
            admin=self.admin,
            contract=self.contract,
            new_code_id=self.new_code_id,
            migrate_msg=bytes(json.dumps(self.migrate_msg), "utf-8"),
        )

    @classmethod
    def from_proto(cls, proto: MsgMigrateContract_pb) -> MsgMigrateContract:
        return cls(
            admin=proto.admin,
            contract=proto.contract,
            new_code_id=proto.new_code_id,
            migrate_msg=proto.migrate_msg,
        )


@attr.s
class MsgUpdateContractAdmin(Msg):
    """Update a smart contract's admin.

    Args:
        admin: address of current admin (sender)
        new_admin: address of new admin
        contract: address of contract to change
    """

    type_amino = "wasm/MsgUpdateContractAdmin"
    """"""
    type_url = "/terra.wasm.v1beta1.MsgUpdateContractAdmin"
    """"""

    admin: AccAddress = attr.ib()
    new_admin: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "admin": self.admin,
                "new_admin": self.new_admin,
                "contract": self.contract
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgUpdateContractAdmin:
        return cls(
            admin=data["admin"],
            new_admin=data["new_admin"],
            contract=data["contract"],
        )

    def to_proto(self) -> MsgUpdateContractAdmin_pb:
        return MsgUpdateContractAdmin_pb(
            admin=self.admin, new_admin=self.new_admin, contract=self.contract
        )

    @classmethod
    def from_proto(cls, proto: MsgUpdateContractAdmin_pb) -> MsgUpdateContractAdmin:
        return cls(
            admin=proto.admin,
            new_admin=proto.new_admin,
            contract=proto.contract,
        )


@attr.s
class MsgClearContractAdmin(Msg):
    """Clears the contract's admin field.

    Args:
        admin: address of current admin (sender)
        contract: address of contract to change
    """

    type_amino = "wasm/MsgClearContractAdmin"
    """"""
    type_url = "/terra.wasm.v1beta1.MsgClearContractAdmin"
    """"""

    admin: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "admin": self.admin,
                "contract": self.contract
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgClearContractAdmin:
        return cls(
            admin=data["admin"],
            contract=data["contract"],
        )

    def to_proto(self) -> MsgClearContractAdmin_pb:
        return MsgClearContractAdmin_pb(admin=self.admin, contract=self.contract)

    @classmethod
    def from_proto(cls, proto: MsgClearContractAdmin_pb) -> MsgClearContractAdmin:
        return cls(
            admin=proto.admin,
            contract=proto.contract,
        )



"""Treasury module data objects."""

from __future__ import annotations

__all__ = ["PolicyConstraints"]

import attr
from terra_proto.terra.treasury.v1beta1 import PolicyConstraints as PolicyConstraints_pb

from terra_sdk.core import Coin, Dec


@attr.s
class PolicyConstraints:
    """Contains information about tax reward or reward weight
    policy constraints.
    """

    rate_min: Dec = attr.ib()
    """"""
    rate_max: Dec = attr.ib()
    """"""
    cap: Coin = attr.ib()
    """"""
    change_rate_max: Dec = attr.ib()
    """"""

    def clamp(self, prev_rate: Dec, new_rate: Dec) -> Dec:
        """Simulates the effect of the policy contraint.

        Args:
            prev_rate (Dec): previous rate
            new_rate (Dec): new rate

        Returns:
            Dec: result of clamp (constrained change)
        """
        prev_rate = Dec(prev_rate)
        new_rate = Dec(new_rate)

        if new_rate < self.rate_min:
            new_rate = self.rate_min
        elif new_rate > self.rate_max:
            new_rate = self.rate_max

        delta = new_rate - prev_rate
        if new_rate > prev_rate:
            if delta > self.change_rate_max:
                new_rate = prev_rate + self.change_rate_max
        else:
            if abs(delta) > self.change_rate_max:
                new_rate = prev_rate - self.change_rate_max
        return new_rate

    def to_amino(self) -> dict:
        return {
            "rate_min": str(self.rate_min),
            "rate_max": str(self.rate_max),
            "cap": self.cap.to_amino(),
            "change_rate_max": str(self.change_rate_max)
        }

    @classmethod
    def from_data(cls, data: dict) -> PolicyConstraints:
        return cls(
            rate_min=Dec(data["rate_min"]),
            rate_max=Dec(data["rate_max"]),
            cap=Coin.from_data(data["cap"]),
            change_rate_max=Dec(data["change_rate_max"]),
        )

    def to_proto(self) -> PolicyConstraints_pb:
        return PolicyConstraints_pb(
            rate_min=str(self.rate_min),
            rate_max=str(self.rate_max),
            cap=self.cap.to_proto(),
            change_rate_max=str(self.change_rate_max),
        )



"""Oracle module messages."""

from __future__ import annotations

import copy
import hashlib

import attr
from terra_proto.terra.oracle.v1beta1 import (
    MsgAggregateExchangeRatePrevote as MsgAggregateExchangeRatePrevote_pb,
)
from terra_proto.terra.oracle.v1beta1 import (
    MsgAggregateExchangeRateVote as MsgAggregateExchangeRateVote_pb,
)
from terra_proto.terra.oracle.v1beta1 import (
    MsgDelegateFeedConsent as MsgDelegateFeedConsent_pb,
)

from terra_sdk.core import AccAddress, Coins, Dec, ValAddress
from terra_sdk.core.msg import Msg
from terra_sdk.util.json import dict_to_data

__all__ = [
    "vote_hash",
    "aggregate_vote_hash",
    "MsgDelegateFeedConsent",
    "MsgAggregateExchangeRatePrevote",
    "MsgAggregateExchangeRateVote",
]


def vote_hash(denom: str, exchange_rate: Dec, salt: str, validator: str) -> str:
    """Calculates vote hash for submitting :class:`MsgExchangeRatePrevote`.

    Args:
        denom (str): denom to vote for
        exchange_rate (Dec): exchange rate of LUNA
        salt (str): salt
        validator (str): validator operator address

    Returns:
        str: vote hash
    """
    payload = f"{denom}:{exchange_rate!s}:{salt}:{validator}"
    sha_hash = hashlib.sha256(payload.encode())
    return sha_hash.hexdigest()[:40]


def aggregate_vote_hash(salt: str, exchange_rates: Coins.Input, validator: str) -> str:
    """Calculates aggregate vote hash for submitting :class:`MsgAggregateExchangeRatePrevote`.

    Args:
        salt (str): salt
        exchange_rates (Coins.Input): exchange rates in various denominations
        validator (str): validator operator address

    Returns:
        str: aggregate vote hash
    """
    payload = f"{salt}:{str(Coins(exchange_rates))}:{validator}"
    sha_hash = hashlib.sha256(payload.encode())
    return sha_hash.hexdigest()[:40]


@attr.s
class MsgDelegateFeedConsent(Msg):
    """Re-assign oracle feeder account for a validator.

    Args:
        operator: validator to change feeder for
        delegate: new feeder address
    """

    type_amino = "oracle/MsgDelegateFeedConsent"
    """"""
    type_url = "/terra.oracle.v1beta1.MsgDelegateFeedConsent"
    """"""
    action = "delegatefeeder"
    """"""

    operator: ValAddress = attr.ib()
    delegate: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "operator": self.operator,
                "delegate": self.delegate
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgDelegateFeedConsent:
        return cls(operator=data["operator"], delegate=data["delegate"])

    def to_proto(self) -> MsgDelegateFeedConsent_pb:
        return MsgDelegateFeedConsent_pb(operator=self.operator, delegate=self.delegate)

    @classmethod
    def from_proto(cls, proto: MsgDelegateFeedConsent_pb) -> MsgDelegateFeedConsent:
        return cls(
            operator=proto.operator,
            delegate=proto.delegate
        )



@attr.s
class MsgAggregateExchangeRatePrevote(Msg):
    """Submit an aggregate vote for the current vote period.

    Args:
        hash: aggregate vote hash
        feeder: account submitting the aggregate prevote
        validator: validator to which the aggregate prevote corresponds
    """

    type_amino = "oracle/MsgAggregateExchangeRatePrevote"
    """"""
    type_url = "/terra.oracle.v1beta1.MsgAggregateExchangeRatePrevote"
    """"""

    hash: str = attr.ib()
    feeder: AccAddress = attr.ib()
    validator: ValAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "hash": self.hash,
                "feeder": self.feeder,
                "validator": self.validator
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgAggregateExchangeRatePrevote:
        return cls(
            hash=data["hash"],
            feeder=data["feeder"],
            validator=data["validator"],
        )

    def to_proto(self) -> MsgAggregateExchangeRatePrevote_pb:
        return MsgAggregateExchangeRatePrevote_pb(
            hash=self.hash, feeder=self.feeder, validator=self.validator
        )

    @classmethod
    def from_proto(cls, proto: MsgAggregateExchangeRatePrevote_pb) -> MsgAggregateExchangeRatePrevote:
        return cls(
            hash=proto.hash,
            feeder=proto.feeder,
            validator=proto.validator,
        )


@attr.s
class MsgAggregateExchangeRateVote(Msg):
    """Submit an aggregate prevote for the current vote.

    Args:
        exchange_rates (Coins.Input): exchange rates to use
        salt: aggregate vote salt
        feeder: feeder account submitting aggregate prevote
        validator: validator vote corresponds to
    """

    type_amino = "oracle/MsgAggregateExchangeRateVote"
    """"""
    type_url = "/terra.oracle.v1beta1.MsgAggregateExchangeRateVote"
    """"""

    exchange_rates: Coins = attr.ib(converter=Coins)
    salt: str = attr.ib()
    feeder: AccAddress = attr.ib()
    validator: ValAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "exchange_rates": str(self.exchange_rates.to_dec_coins()),
                "salt": self.salt,
                "feeder": self.feeder,
                "validator": self.validator
            }
        }

    def to_data(self) -> dict:
        d = copy.deepcopy(self.__dict__)
        d["exchange_rates"] = str(self.exchange_rates.to_dec_coins())
        return {"type": self.type, "value": dict_to_data(d)}

    @classmethod
    def from_data(cls, data: dict) -> MsgAggregateExchangeRateVote:
        return cls(
            exchange_rates=Coins.from_str(data["exchange_rates"]),
            salt=data["salt"],
            feeder=data["feeder"],
            validator=data["validator"],
        )

    def to_proto(self) -> MsgAggregateExchangeRateVote_pb:
        return MsgAggregateExchangeRateVote_pb(
            exchange_rates=str(self.exchange_rates),
            salt=self.salt,
            feeder=self.feeder,
            validator=self.validator,
        )

    @classmethod
    def from_proto(cls, proto: MsgAggregateExchangeRateVote_pb) -> MsgAggregateExchangeRateVote:
        return cls(
            exchange_rates=Coins.from_proto(proto.exchange_rates),
            salt=proto.salt,
            feeder=proto.feeder,
            validator=proto.validator,
        )

    def get_aggregate_vote_hash(self) -> str:
        """Vote hash required for message's associated prevote.

        Returns:
            str: aggregate vote hash
        """
        return aggregate_vote_hash(self.salt, self.exchange_rates, self.validator)

    def get_aggregate_prevote(self) -> MsgAggregateExchangeRatePrevote:
        """Generates the associated :class:`MsgAggregateExchangeRatePrevote` object with
        the correct prepopulated fields.

        Returns:
            MsgAggregateExchangeRatePrevote: associated aggregate prevote
        """
        return MsgAggregateExchangeRatePrevote(
            hash=self.get_aggregate_vote_hash(),
            feeder=self.feeder,
            validator=self.validator,
        )



"""Some useful base classes to inherit from."""
from abc import abstractmethod
from typing import Any, Callable, Dict, List

from betterproto import Message

from .json import JSONSerializable, dict_to_data


class BaseTerraData(JSONSerializable, Message):

    type: str
    type_url: str

    def to_data(self) -> dict:
        return {"type": self.type_url, "value": dict_to_data(self.__dict__)}

    @abstractmethod
    def to_proto(self):
        pass


def create_demux(inputs: List) -> Callable[[Dict[str, Any]], Any]:
    table = {i.type_url: i.from_data for i in inputs}

    def from_data(data: dict):
        return table[data["@type"]](data)

    return from_data


def create_demux_proto(inputs: List) -> Callable[[Dict[str, Any]], Any]:
    table = {i.type_url: i.from_proto for i in inputs}

    def from_proto(data: dict):
        return table[data["@type"]](data)

    return from_proto



from .client import MsgCreateClient, MsgUpdateClient, MsgUpgradeClient, MsgSubmitMisbehaviour
from .connection import MsgConnectionOpenInit, MsgConnectionOpenTry, MsgConnectionOpenAck, MsgConnectionOpenConfirm
from .channel import (
    MsgChannelOpenInit, MsgChannelOpenTry, MsgChannelOpenAck, MsgChannelOpenConfirm,
    MsgChannelCloseInit, MsgChannelCloseConfirm,
    MsgRecvPacket, MsgTimeout, MsgAcknowledgement
)

__all__ = [
    "MsgCreateClient",
    "MsgUpdateClient",
    "MsgUpgradeClient",
    "MsgSubmitMisbehaviour",
    "MsgConnectionOpenInit",
    "MsgConnectionOpenTry",
    "MsgConnectionOpenAck",
    "MsgConnectionOpenConfirm",
    "MsgChannelOpenInit",
    "MsgChannelOpenTry",
    "MsgChannelOpenAck",
    "MsgChannelOpenConfirm",
    "MsgChannelCloseInit",
    "MsgChannelCloseConfirm",
    "MsgRecvPacket",
    "MsgTimeout",
    "MsgAcknowledgement"
]


from .data import AggregateExchangeRatePrevote, AggregateExchangeRateVote
from .msgs import (
    MsgAggregateExchangeRatePrevote,
    MsgAggregateExchangeRateVote,
    MsgDelegateFeedConsent,
    aggregate_vote_hash,
    vote_hash,
)

__all__ = [
    "AggregateExchangeRatePrevote",
    "AggregateExchangeRateVote",
    "vote_hash",
    "aggregate_vote_hash",
    "MsgDelegateFeedConsent",
    "MsgAggregateExchangeRatePrevote",
    "MsgAggregateExchangeRateVote",
]



"""ibc client module data objects."""
from __future__ import annotations

from typing import List

import attr
from terra_proto.ibc.core.client.v1 import (
    Height as Height_pb,
    IdentifiedClientState as IdentifiedClientState_pb,
    ClientConsensusStates as ClientConsensusStates_pb,
    ConsensusStateWithHeight as ConsensusStateWithHeight_pb,
    Params as Params_pb
)
from betterproto.lib.google.protobuf import Any as Any_pb

from terra_sdk.util.json import JSONSerializable

__all__ = ["Height", "IdentifiedClientState", "ConsensusStateWithHeight", "ClientConsensusStates", "Params"]


@attr.s
class Height(JSONSerializable):
    revision_number: int = attr.ib(converter=int)
    revision_height: int = attr.ib(converter=int)

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Height:
        return cls(
            revision_number=data["revision_number"],
            revision_height=data["revision_height"],
        )

    def to_proto(self) -> Height_pb:
        return Height_pb(
            revision_number=self.revision_number, revision_height=self.revision_height
        )

    @classmethod
    def from_proto(cls, proto: Height_pb) -> Height:
        return cls(
            revision_number=proto.revision_number,
            revision_height=proto.revision_height,
        )

@attr.s
class IdentifiedClientState(JSONSerializable):
    """
    IdentifiedClientState defines a client state with an additional client identifier field.
    """

    client_id: str = attr.ib()
    client_state: dict = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> IdentifiedClientState:
        return cls(
            client_id=data["client_id"],
            consensus_state=Any_pb().from_dict(data["client_state"])
        )

    def to_proto(self) -> IdentifiedClientState_pb:
        return IdentifiedClientState_pb(
            client_id=self.client_id,
            client_state=Any_pb().from_dict(self.consensus_state)
        )

    @classmethod
    def from_proto(cls, proto: IdentifiedClientState_pb) -> IdentifiedClientState:
        return cls(
            client_id=proto.client_id,
            consensus_state=proto.client_state.to_dict()
        )


@attr.s
class ConsensusStateWithHeight(JSONSerializable):
    """
    ConsensusStateWithHeight defines a consensus state with an additional height field.
    """

    height: str = attr.ib()
    consensus_state: dict = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> ConsensusStateWithHeight:
        return cls(
            height=data["height"],
            consensus_state=Any_pb().from_dict(data["consensus_state"])
        )

    def to_proto(self) -> ConsensusStateWithHeight_pb:
        return ConsensusStateWithHeight_pb(
            height=self.height,
            consensus_state=Any_pb().from_dict(self.consensus_state)
        )

    @classmethod
    def from_proto(cls, proto: ConsensusStateWithHeight_pb) -> ConsensusStateWithHeight:
        return cls(
            height=proto.height,
            consensus_state=proto.consensus_state.to_dict()
        )


@attr.s
class ClientConsensusStates(JSONSerializable):
    """
    ClientConsensusStates defines all the stored consensus states for a given client.
    """

    client_id: str = attr.ib()
    consensus_states: List[ConsensusStateWithHeight] = attr.ib(converter=list)

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> ClientConsensusStates:
        return cls(
            client_id=data["client_id"],
            consensus_states=[ConsensusStateWithHeight.from_data(state) for state in data["consensus_states"]]
        )

    def to_proto(self) -> ClientConsensusStates_pb:
        return ClientConsensusStates_pb(
            client_id=self.client_id,
            consensus_states=[state.to_proto for state in self.consensus_states]
        )

    @classmethod
    def from_proto(cls, proto: ClientConsensusStates_pb) -> ClientConsensusStates:
        return cls(
            client_id=proto.client_id,
            consensus_states=[ConsensusStateWithHeight.from_proto(state) for state in proto.consensus_states]
        )



@attr.s
class Params(JSONSerializable):
    """
    Params defines the set of IBC light client parameters.
    """

    allowed_clients: List[str] = attr.ib(converter=list)

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Params:
        return cls(
            allowed_clients=data["allowed_clients"]
        )

    def to_proto(self) -> Params_pb:
        return Params_pb(
            allowed_clients=self.allowed_clients
        )

    @classmethod
    def from_proto(cls, proto: Params_pb) -> Params:
        return cls(
            allowed_clients=proto.allowed_clients
        )




from datetime import datetime


def to_isoformat(dt: datetime) -> str:
    return dt.isoformat(timespec="milliseconds").replace("+00:00", "Z").replace("000Z", "Z")


def replace_bytes_index(text: bytes, index = 0, replacement: bytes = ''):
    return '%s%s%s'%(text[:index],replacement,text[index+1:])



from terra_sdk.core.distribution.proposals import CommunityPoolSpendProposal
from terra_sdk.core.gov.proposals import TextProposal
from terra_sdk.core.params.proposals import ParameterChangeProposal

from .base import create_demux

parse_content = create_demux(
    [
        CommunityPoolSpendProposal,
        TextProposal,
        ParameterChangeProposal,
    ]
)



"""Data objects pertaining to accounts."""

from __future__ import annotations

from typing import List, Optional

import attr
from terra_proto.terra.vesting.v1beta1 import (
    LazyGradedVestingAccount as LazyGradedVestingAccount_pb,
)

from terra_sdk.core import AccAddress, Coins
from terra_sdk.util.json import JSONSerializable

from ...public_key import PublicKey
from .base_account import BaseAccount

__all__ = ["LazyGradedVestingAccount"]


@attr.s
class LazyGradedVestingAccount(BaseAccount):
    """Stores information about an account with vesting."""

    address: AccAddress = attr.ib()
    """"""

    public_key: PublicKey = attr.ib()
    """"""

    account_number: int = attr.ib(converter=int)
    """"""

    sequence: int = attr.ib(converter=int)
    """"""

    original_vesting: Coins = attr.ib(converter=Coins)
    """"""

    delegated_free: Coins = attr.ib(converter=Coins)
    """"""

    delegated_vesting: Coins = attr.ib(converter=Coins)
    """"""

    end_time: int = attr.ib(converter=int)
    """"""

    vesting_schedules: List[dict] = attr.ib()
    """"""

    type_amino = "core/LazyGradedVestingAccount"
    type_url = "/terra.vesting.v1beta1.LazyGradedVestingAccount"

    def get_sequence(self) -> int:
        return self.sequence

    def get_account_number(self) -> int:
        return self.account_number

    def get_public_key(self) -> PublicKey:
        return self.public_key

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "address": self.address,
                "public_key": self.public_key and self.public_key.to_amino(),
                "account_number": str(self.account_number),
                "sequence": str(self.sequence),
                "original_vesting": self.original_vesting.to_amino(),
                "delegated_free": self.delegated_free.to_amino(),
                "delegated_vesting": self.delegated_vesting.to_amino(),
                "end_time": str(self.end_time),
                "vesting_schedules": self.vesting_schedules,
            }
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "address": self.address,
            "public_key": self.public_key and self.public_key.to_data(),
            "account_number": str(self.account_number),
            "sequence": str(self.sequence),
            "original_vesting": self.original_vesting.to_data(),
            "delegated_free": self.delegated_free.to_data(),
            "delegated_vesting": self.delegated_vesting.to_data(),
            "end_time": str(self.end_time),
            "vesting_schedules": self.vesting_schedules,
        }

    @classmethod
    def from_data(cls, data: dict) -> LazyGradedVestingAccount:
        return cls(
            address=data["address"],
            public_key=PublicKey.from_data(data["public_key"]),
            account_number=data["account_number"],
            sequence=data["sequence"],
            original_vesting=Coins.from_data(data["original_vesting"]),
            delegated_free=Coins.from_data(data["delegated_free"]),
            delegated_vesting=Coins.from_data(data["delegated_vesting"]),
            end_time=data["end_time"],
            vesting_schedules=data["vesting_schedules"],
        )

    def to_proto(self) -> LazyGradedVestingAccount_pb:
        proto = LazyGradedVestingAccount_pb()
        proto.base_vesting_account = self.base_vesting_account.to_proto()
        proto.vesting_schedules = [vs.to_proto() for vs in self.vesting_schedules]
        return proto



from .data import (
    Account,
    LazyGradedVestingAccount,
    PublicKey,
    TxInfo,
    TxLog,
    parse_tx_logs,
)

__all__ = [
    "Account",
    "LazyGradedVestingAccount",
    "TxLog",
    "TxInfo",
    "PublicKey",
    "parse_tx_logs",
]



from terra_sdk.core import Coin, Dec

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncMarketAPI", "MarketAPI"]


class AsyncMarketAPI(BaseAsyncAPI):
    async def swap_rate(self, offer_coin: Coin, ask_denom: str) -> Coin:
        """Simulates a swap given an amount offered and a target denom.

        Args:
            offer_coin (Coin): amount offered (swap from)
            ask_denom (str): target denom (swap to)

        Returns:
            Coin: simulated amount received
        """
        params = {"offer_coin": str(offer_coin), "ask_denom": ask_denom}
        res = await self._c._get("/terra/market/v1beta1/swap", params)
        return Coin.from_data(res.get("return_coin"))

    async def terra_pool_delta(self) -> Dec:
        """Fetches the Terra pool delta.

        Returns:
            Dec: Terra pool delta
        """
        res = await self._c._get("/terra/market/v1beta1/terra_pool_delta")
        return Dec(res.get("terra_pool_delta"))

    async def parameters(self) -> dict:
        """Fetches the Market module's parameters.

        Returns:
            dict: Market module parameters
        """
        res = await self._c._get("/terra/market/v1beta1/params")
        params = res["params"]
        return {
            "base_pool": Dec(params.get("base_pool")),
            "pool_recovery_period": int(params.get("pool_recovery_period")),
            "min_stability_spread": Dec(params.get("min_stability_spread")),
        }


class MarketAPI(AsyncMarketAPI):
    @sync_bind(AsyncMarketAPI.swap_rate)
    def swap_rate(self, offer_coin: Coin, ask_denom: str) -> Coin:
        pass

    swap_rate.__doc__ = AsyncMarketAPI.swap_rate.__doc__

    @sync_bind(AsyncMarketAPI.terra_pool_delta)
    def terra_pool_delta(self) -> Dec:
        pass

    terra_pool_delta.__doc__ = AsyncMarketAPI.terra_pool_delta.__doc__

    @sync_bind(AsyncMarketAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncMarketAPI.parameters.__doc__



from .msgs import (
    MsgClearContractAdmin,
    MsgExecuteContract,
    MsgInstantiateContract,
    MsgMigrateCode,
    MsgMigrateContract,
    MsgStoreCode,
    MsgUpdateContractAdmin,
)

__all__ = [
    "MsgStoreCode",
    "MsgMigrateCode",
    "MsgInstantiateContract",
    "MsgExecuteContract",
    "MsgMigrateContract",
    "MsgUpdateContractAdmin",
    "MsgClearContractAdmin",
]



"""Market module message types."""

from __future__ import annotations

import attr
from terra_proto.terra.market.v1beta1 import MsgSwap as MsgSwap_pb
from terra_proto.terra.market.v1beta1 import MsgSwapSend as MsgSwapSend_pb

from terra_sdk.core import AccAddress, Coin
from terra_sdk.core.msg import Msg

__all__ = ["MsgSwap", "MsgSwapSend"]


@attr.s
class MsgSwap(Msg):
    """Perform a native on-chain swap from ``offer_coin`` to ``ask_denom``.

    Args:
        trader: account performing swap
        offer_coin (Union[Coin, str, dict]): coin offered for swap
        ask_denom: denom into which to swap
    """

    type_amino = "market/MsgSwap"
    """"""
    type_url = "/terra.market.v1beta1.MsgSwap"
    """"""
    action = "swap"
    """"""

    trader: AccAddress = attr.ib()
    offer_coin: Coin = attr.ib(converter=Coin.parse)  # type: ignore
    ask_denom: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "trader": self.trader,
                "offer_coin": self.offer_coin.to_amino(),
                "ask_denom": self.ask_denom
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgSwap:
        return cls(
            trader=data["trader"],
            offer_coin=Coin.from_data(data["offer_coin"]),
            ask_denom=data["ask_denom"],
        )

    def to_proto(self) -> MsgSwap_pb:
        return MsgSwap_pb(
            trader=self.trader,
            offer_coin=self.offer_coin.to_proto(),
            ask_denom=self.ask_denom,
        )

    @classmethod
    def from_proto(cls, proto: MsgSwap_pb) -> MsgSwap:
        return cls(
            trader=proto.trader,
            offer_coin=Coin.from_proto(proto.offer_coin),
            ask_denom=proto.ask_denom,
        )


@attr.s
class MsgSwapSend(Msg):
    """Performs a swap and sends the resultant swapped amount to ``to_address``.

    Args:
        from_address: account performing swap
        to_address: account which will received resultant funds from swap
        offer_coin (Union[Coin, str, dict]): coin offered for swap
        ask_denom: denom into which to swap
    """

    type_amino = "market/MsgSwapSend"
    """"""
    type_url = "/terra.market.v1beta1.MsgSwapSend"
    """"""
    action = "swapsend"
    """"""

    from_address: AccAddress = attr.ib()
    to_address: AccAddress = attr.ib()
    offer_coin: Coin = attr.ib(converter=Coin.parse)  # type: ignore
    ask_denom: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "from_address": self.from_address,
                "to_address": self.to_address,
                "offer_coin": self.offer_coin.to_amino(),
                "ask_denom": self.ask_denom
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgSwapSend:
        return cls(
            from_address=data["from_address"],
            to_address=data["to_address"],
            offer_coin=Coin.from_data(data["offer_coin"]),
            ask_denom=data["ask_denom"],
        )

    def to_proto(self) -> MsgSwapSend_pb:
        return MsgSwapSend_pb(
            from_address=self.from_address,
            to_address=self.to_address,
            offer_coin=self.offer_coin.to_proto(),
            ask_denom=self.ask_denom,
        )

    @classmethod
    def from_proto(cls, proto: MsgSwapSend_pb) -> MsgSwapSend:
        return cls(
            from_address=proto.from_address,
            to_address=proto.to_address,
            offer_coin=Coin.from_proto(proto.offer_coin),
            ask_denom=proto.ask_denom,
        )




from typing import List, Optional

from terra_sdk.core import AccAddress, Coin, Coins, Dec, Numeric, ValAddress
from terra_sdk.core.oracle import (
    AggregateExchangeRatePrevote,
    AggregateExchangeRateVote,
)
from terra_sdk.exceptions import LCDResponseError

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncOracleAPI", "OracleAPI"]


class AsyncOracleAPI(BaseAsyncAPI):
    async def exchange_rates(self) -> Coins:
        """Fetches registered exchange rates of Luna in all available denoms.

        Returns:
            Coins: exchange rates of Luna
        """
        res = await self._c._get("/terra/oracle/v1beta1/denoms/exchange_rates")
        rates = res.get("exchange_rates")
        if res:
            return Coins.from_data(rates)
        else:
            return Coins({})

    async def exchange_rate(self, denom: str) -> Coin:
        """Fetches registered exchange rate of Luna in a specific denom.

        Args:
            denom (str): denom

        Returns:
            Coin: exchange rate of Luna
        """
        res = await self._c._get(f"/terra/oracle/v1beta1/denoms/{denom}/exchange_rate")
        return Dec(res.get("exchange_rate"))

    async def active_denoms(self) -> List[str]:
        """Fetches current active denoms.

        Returns:
            List[str]: active denoms
        """
        res = await self._c._get("/terra/oracle/v1beta1/denoms/actives")
        return res.get("actives")

    async def feeder_address(self, validator: ValAddress) -> AccAddress:
        """Fetches associated feeder address for a validator.

        Args:
            validator (ValAddress): validator operator address

        Returns:
            AccAddress: feeder address
        """
        res = await self._c._get(f"/terra/oracle/v1beta1/validators/{validator}/feeder")
        return res.get("feeder_addr")

    async def misses(self, validator: ValAddress) -> int:
        """Fetches current value of miss counter for a validator.

        Args:
            validator (ValAddress): validator operator address

        Returns:
            int: current number of misses
        """
        res = await self._c._get(f"terra/oracle/v1beta1/validators/{validator}/miss")
        return int(res.get("miss_counter"))

    async def aggregate_prevote(
        self, validator: ValAddress
    ) -> Optional[AggregateExchangeRatePrevote]:
        """Fetches active aggregate prevote for a validator.

        Args:
            validator (ValAddress): validator operator address

        Returns:
            Optional[AggregateExchangeRatePrevote]: current aggegate prevote (if any).
        """
        try:
            res = await self._c._get(
                f"/terra/oracle/v1beta1/validators/{validator}/aggregate_prevote"
            )
        except LCDResponseError as e:
            if e.response.status == 404:
                return None
            else:
                raise e
        return AggregateExchangeRatePrevote.from_data(res.get("aggregate_prevote"))

    async def aggregate_vote(
        self, validator: ValAddress
    ) -> Optional[AggregateExchangeRateVote]:
        """Fetches active aggregate vote for a validator.

        Args:
            validator (ValAddress): validator operator address

        Returns:
            Optional[AggregateExchangeRatePrevote]: current aggegate vote (if any).
        """
        try:
            res = await self._c._get(
                f"/terra/oracle/v1beta1/validators/{validator}/aggregate_vote"
            )
        except LCDResponseError as e:
            if e.response.status == 404:
                return None
            else:
                raise e
        return AggregateExchangeRateVote.from_data(res.get("aggregate_vote"))

    async def parameters(self) -> dict:
        """Fetches Oracle module parameters.

        Returns:
            dict: Oracle module parameters
        """
        res = await self._c._get("/terra/oracle/v1beta1/params")
        params = res.get("params")
        return {
            "vote_period": Numeric.parse(params["vote_period"]),
            "vote_threshold": Dec(params["vote_threshold"]),
            "reward_band": Dec(params["reward_band"]),
            "reward_distribution_window": Numeric.parse(
                params["reward_distribution_window"]
            ),
            "whitelist": [
                {"name": x["name"], "tobin_tax": Dec(x["tobin_tax"])}
                for x in params["whitelist"]
            ],
            "slash_fraction": Dec(params["slash_fraction"]),
            "slash_window": Numeric.parse(params["slash_window"]),
            "min_valid_per_window": Dec(params["min_valid_per_window"]),
        }


class OracleAPI(AsyncOracleAPI):
    @sync_bind(AsyncOracleAPI.exchange_rates)
    def exchange_rates(self) -> Coins:
        pass

    exchange_rates.__doc__ = AsyncOracleAPI.exchange_rates.__doc__

    @sync_bind(AsyncOracleAPI.exchange_rate)
    def exchange_rate(self, denom: str) -> Coin:
        pass

    exchange_rate.__doc__ = AsyncOracleAPI.exchange_rate.__doc__

    @sync_bind(AsyncOracleAPI.active_denoms)
    def active_denoms(self) -> List[str]:
        pass

    active_denoms.__doc__ = AsyncOracleAPI.active_denoms.__doc__

    @sync_bind(AsyncOracleAPI.feeder_address)
    def feeder_address(self, validator: ValAddress) -> AccAddress:
        pass

    feeder_address.__doc__ = AsyncOracleAPI.active_denoms.__doc__

    @sync_bind(AsyncOracleAPI.misses)
    def misses(self, validator: ValAddress) -> int:
        pass

    misses.__doc__ = AsyncOracleAPI.misses.__doc__

    @sync_bind(AsyncOracleAPI.aggregate_prevote)
    def aggregate_prevote(
        self, validator: ValAddress
    ) -> Optional[AggregateExchangeRatePrevote]:
        pass

    aggregate_prevote.__doc__ = AsyncOracleAPI.aggregate_prevote.__doc__

    @sync_bind(AsyncOracleAPI.aggregate_vote)
    def aggregate_vote(
        self, validator: ValAddress
    ) -> Optional[AggregateExchangeRateVote]:
        pass

    aggregate_vote.__doc__ = AsyncOracleAPI.aggregate_vote.__doc__

    @sync_bind(AsyncOracleAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncOracleAPI.parameters.__doc__



from terra_sdk.core.authz import (
    MsgExecAuthorized,
    MsgGrantAuthorization,
    MsgRevokeAuthorization,
)
from terra_sdk.core.bank import MsgMultiSend, MsgSend
from terra_sdk.core.distribution import (
    MsgFundCommunityPool,
    MsgSetWithdrawAddress,
    MsgWithdrawDelegationReward,
    MsgWithdrawValidatorCommission,
)
from terra_sdk.core.gov.msgs import MsgDeposit, MsgSubmitProposal, MsgVote
from terra_sdk.core.market import MsgSwap, MsgSwapSend
from terra_sdk.core.oracle import (
    MsgAggregateExchangeRatePrevote,
    MsgAggregateExchangeRateVote,
    MsgDelegateFeedConsent,
)
from terra_sdk.core.slashing import MsgUnjail
from terra_sdk.core.staking import (
    MsgBeginRedelegate,
    MsgCreateValidator,
    MsgDelegate,
    MsgEditValidator,
    MsgUndelegate,
)
from terra_sdk.core.wasm import (
    MsgClearContractAdmin,
    MsgExecuteContract,
    MsgInstantiateContract,
    MsgMigrateCode,
    MsgMigrateContract,
    MsgStoreCode,
    MsgUpdateContractAdmin,
)
from terra_sdk.core.ibc.msgs import (
    MsgCreateClient,
    MsgUpdateClient,
    MsgUpgradeClient,
    MsgSubmitMisbehaviour,
    MsgConnectionOpenInit,
    MsgConnectionOpenTry,
    MsgConnectionOpenAck,
    MsgConnectionOpenConfirm,
    MsgChannelOpenInit,
    MsgChannelOpenTry,
    MsgChannelOpenAck,
    MsgChannelOpenConfirm,
    MsgChannelCloseInit,
    MsgChannelCloseConfirm,
    MsgRecvPacket,
    MsgTimeout,
    MsgAcknowledgement
)
from terra_sdk.core.ibc_transfer import (
    MsgTransfer
)

from .base import create_demux, create_demux_proto

bank_msgs = [MsgSend, MsgMultiSend]
distribution_msgs = [
    MsgFundCommunityPool,
    MsgSetWithdrawAddress,
    MsgWithdrawDelegationReward,
    MsgWithdrawValidatorCommission,
]
gov_msgs = [MsgDeposit, MsgSubmitProposal, MsgVote]
market_msgs = [MsgSwap, MsgSwapSend]
authz_msgs = [
    MsgExecAuthorized,
    MsgGrantAuthorization,
    MsgRevokeAuthorization,
]
oracle_msgs = [
    MsgAggregateExchangeRatePrevote,
    MsgAggregateExchangeRateVote,
    MsgDelegateFeedConsent,
]
slashing_msgs = [MsgUnjail]
staking_msgs = [
    MsgBeginRedelegate,
    MsgCreateValidator,
    MsgDelegate,
    MsgEditValidator,
    MsgUndelegate,
]
wasm_msgs = [
    MsgStoreCode,
    MsgMigrateCode,
    MsgInstantiateContract,
    MsgExecuteContract,
    MsgMigrateContract,
    MsgUpdateContractAdmin,
    MsgClearContractAdmin,
]

ibc_transfer_msgs = [
    MsgTransfer
]
ibc_msgs = [
    MsgCreateClient,
    MsgUpdateClient,
    MsgUpgradeClient,
    MsgSubmitMisbehaviour,
    MsgConnectionOpenInit,
    MsgConnectionOpenTry,
    MsgConnectionOpenAck,
    MsgConnectionOpenConfirm,
    MsgChannelOpenInit,
    MsgChannelOpenTry,
    MsgChannelOpenAck,
    MsgChannelOpenConfirm,
    MsgChannelCloseInit,
    MsgChannelCloseConfirm,
    MsgRecvPacket,
    MsgTimeout,
    MsgAcknowledgement
]

parse_msg = create_demux(
    [
        *bank_msgs,
        *distribution_msgs,
        *gov_msgs,
        *market_msgs,
        *oracle_msgs,
        *slashing_msgs,
        *staking_msgs,
        *wasm_msgs,
        *ibc_msgs,
        *ibc_transfer_msgs
    ]
)
parse_proto = create_demux_proto(
    [
        *bank_msgs,
        *distribution_msgs,
        *gov_msgs,
        *market_msgs,
        *oracle_msgs,
        *slashing_msgs,
        *staking_msgs,
        *wasm_msgs,
        *ibc_msgs,
        *ibc_transfer_msgs
    ]
)


"""ibc connection module message types."""

from __future__ import annotations

from typing import Optional, Any, List

import attr
from terra_proto.ibc.core.client.v1 import (
    MsgCreateClient as MsgCreateClient_pb,
    MsgUpdateClient as MsgUpdateClient_pb,
    MsgUpgradeClient as MsgUpgradeClient_pb,
    MsgSubmitMisbehaviour as MsgSubmitMisbehaviour_pb
)
from betterproto.lib.google.protobuf import Any as Any_pb

from terra_sdk.core.msg import Msg

__all__ = ["MsgCreateClient", "MsgUpdateClient", "MsgUpgradeClient", "MsgSubmitMisbehaviour"]


@attr.s
class MsgCreateClient(Msg):
    """
    MsgCreateClientResponse defines the Msg/CreateClient response type.
    """

    type_url = "/ibc.core.client.v1.MsgCreateClient"
    """"""

    client_state: dict = attr.ib()
    consensus_state: dict = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgCreateClient:
        return cls(
            client_state=data["client_state"],
            consensus_state=data["consensus_state"],
            signer=data["signer"]
        )

    def to_proto(self) -> MsgCreateClient_pb:
        return MsgCreateClient_pb(
            client_state=Any_pb().from_dict(self.client_state),
            consensus_state=Any_pb().from_dict(self.consensus_state),
            signer=self.signer
        )

    @classmethod
    def from_proto(cls, proto: MsgCreateClient_pb) -> MsgCreateClient:
        return cls(
            client_state=proto.client_state.to_dict(),
            consensus_state=proto.consensus_state.to_dict(),
            signer=proto.signer
        )


@attr.s
class MsgUpdateClient(Msg):
    """
    MsgUpdateClient defines an sdk.Msg to update a IBC client state using the given header.
    """

    type_url = "/ibc.core.client.v1.MsgUpdateClient"
    """"""

    client_id: str = attr.ib()
    header: dict = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgUpdateClient:
        return cls(
            client_id=data["client_id"],
            header=data["header"],
            signer=data["signer"]
        )

    def to_proto(self) -> MsgUpdateClient_pb:
        return MsgUpdateClient_pb(
            client_id=self.client_id,
            header=Any_pb().from_dict(self.header),
            signer=self.signer
        )

    @classmethod
    def from_proto(cls, proto: MsgUpdateClient_pb) -> MsgUpdateClient:
        return cls(
            client_id=proto.client_id,
            header=proto.header.to_dict(),
            signer=proto.signer
        )



@attr.s
class MsgUpgradeClient(Msg):
    """
    MsgUpgradeClient defines an sdk.Msg to upgrade an IBC client to a new client state
    """

    type_url = "/ibc.core.client.v1.MsgUpgradeClient"
    """"""

    client_id: str = attr.ib()
    client_state: dict = attr.ib()
    consensus_state: dict = attr.ib()
    proof_upgrade_client: bytes = attr.ib()
    proof_upgrade_consensus_state: bytes = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgUpgradeClient:
        return cls(
            client_id=data["client_id"],
            client_state=data["client_state"],
            consensus_state=data["consensus_state"],
            proof_upgrade_client=data["proof_upgrade_client"],
            proof_upgrade_consensus_state=data["proof_upgrade_consensus_state"],
            signer=data["signer"]
        )

    def to_proto(self) -> MsgUpgradeClient_pb:
        return MsgUpgradeClient_pb(
            client_id=self.client_id,
            client_state=Any_pb().from_dict(self.client_state),
            consensus_state=Any_pb().from_dict(self.consensus_state),
            proof_upgrade_client=self.proof_upgrade_client,
            proof_upgrade_consensus_state=self.proof_upgrade_consensus_state,
            signer=self.signer
        )

    @classmethod
    def from_proto(cls, proto: MsgUpgradeClient_pb) -> MsgUpgradeClient:
        return cls(
            client_id=proto.client_id,
            client_state=proto.client_state.to_dict(),
            consensus_state=proto.consensus_state.to_dict(),
            proof_upgrade_client=proto.proof_upgrade_client,
            proof_upgrade_consensus_state=proto.proof_upgrade_consensus_state,
            signer=proto.signer
        )


@attr.s
class MsgSubmitMisbehaviour(Msg):
    """
    MsgSubmitMisbehaviour defines an sdk.Msg type that submits Evidence for light client misbehaviour.
    """

    type_url = "/ibc.core.client.v1.MsgSubmitMisbehaviour"
    """"""

    client_id: str = attr.ib()
    misbehaviour: dict = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgSubmitMisbehaviour:
        return cls(
            client_id=data["client_id"],
            misbehaviour=data["misbehaviour"],
            signer=data["signer"]
        )

    def to_proto(self) -> MsgSubmitMisbehaviour_pb:
        return MsgSubmitMisbehaviour_pb(
            client_id=self.client_id,
            misbehaviour=Any_pb().from_dict(self.misbehaviour),
            signer=self.signer
        )

    @classmethod
    def from_data(cls, data: dict) -> MsgSubmitMisbehaviour:
        return cls(
            client_id=data["client_id"],
            misbehaviour=data["misbehaviour"],
            signer=data["signer"]
        )

    @classmethod
    def from_proto(cls, proto: MsgSubmitMisbehaviour_pb) -> MsgSubmitMisbehaviour:
        return cls(
            client_id=proto["client_id"],
            misbehaviour=Any_pb().from_dict(proto["misbehaviour"]),
            signer=proto["signer"]
        )





"""feegrant module data objects."""
from __future__ import annotations

import datetime
from typing import List, Union, Optional

import attr
from attr import converters
from dateutil import parser
from terra_proto.cosmos.feegrant.v1beta1 import (
    AllowedMsgAllowance as AllowedMsgAllowance_pb,
)
from terra_proto.cosmos.feegrant.v1beta1 import BasicAllowance as BasicAllowance_pb
from terra_proto.cosmos.feegrant.v1beta1 import (
    PeriodicAllowance as PeriodicAllowance_pb,
)

from terra_sdk.core import Coins
from terra_sdk.util.converter import to_isoformat
from terra_sdk.util.json import JSONSerializable

__all__ = ["BasicAllowance", "PeriodicAllowance", "AllowedMsgAllowance", "Allowance"]


@attr.s
class BasicAllowance(JSONSerializable):
    """
    BasicAllowance implements Allowance with a one-time grant of tokens
    that optionally expires. The grantee can use up to SpendLimit to cover fees.
    """

    spend_limit: Optional[Coins] = attr.ib(converter=converters.optional(Coins))
    expiration: Optional[datetime] = attr.ib(converter=converters.optional(parser.parse))

    type_amino = "feegrant/BasicAllowance"
    type_url = "/cosmos.feegrant.v1beta1.BasicAllowance"

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "spend_limit": self.spend_limit.to_amino() if self.spend_limit else None,
                "expiration": to_isoformat(self.expiration) if self.expiration else None
            }
        }

    def to_data(self) -> dict:
        return {
            "spend_limit": self.spend_limit.to_data() if self.spend_limit else None,
            "expiration": to_isoformat(self.expiration) if self.expiration else None
        }

    @classmethod
    def from_data(cls, data: dict) -> BasicAllowance:
        sl = data.get("spend_limit")
        exp = data.get("expiration")
        return cls(
            spend_limit=Coins.from_data(sl) if sl else None,
            expiration=exp if exp else None
        )

    def to_proto(self) -> BasicAllowance_pb:
        return BasicAllowance_pb(
            spend_limit=self.spend_limit.to_proto() if self.spend_limit else [],
            expiration=self.expiration
        )


@attr.s
class PeriodicAllowance(JSONSerializable):
    """
    PeriodicAllowance extends Allowance to allow for both a maximum cap,
     as well as a limit per time period.
    """

    basic: BasicAllowance = attr.ib()
    period: int = attr.ib(converter=int)
    period_spend_limit: Coins = attr.ib(converter=Coins)
    period_can_spend: Coins = attr.ib(converter=Coins)
    period_reset: datetime = attr.ib(converter=parser.parse)

    type_amino = "feegrant/PeriodicAllowance"
    """"""
    type_url = "/cosmos.feegrant.v1beta1.PeriodicAllowance"
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "basic": self.basic.to_amino(),
                "period": str(self.period),
                "period_spend_limit": self.period_spend_limit.to_amino(),
                "period_can_spend": self.period_can_spend.to_amino(),
                "period_reset": to_isoformat(self.period_reset)
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> PeriodicAllowance:
        return cls(
            basic=BasicAllowance.from_data(data["basic"]),
            period=data["period"],
            period_spend_limit=Coins.from_data(data["period_spend_limit"]),
            period_can_spend=Coins.from_data(data["period_can_spend"]),
            period_reset=data["period_reset"],
        )

    def to_proto(self) -> PeriodicAllowance_pb:
        return PeriodicAllowance_pb(
            basic=self.basic.to_proto(),
            period=self.period,
            period_spend_limit=self.period_spend_limit.to_proto(),
            period_can_spend=self.period_can_spend.to_proto(),
            period_reset=self.period_reset,
        )


@attr.s
class AllowedMsgAllowance(JSONSerializable):
    """
    AllowedMsgAllowance creates allowance only for specified message types.
    """

    allowance: Allowance = attr.ib()
    allowed_messages: List[str] = attr.ib(converter=list)

    type_amino = "feegrant/AllowedMsgAllowance"
    """"""
    type_url = "/cosmos.feegrant.v1beta1.AllowedMsgAllowance"
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "allowance": self.allowance.to_amino(),
                "allowed_messages": self.allowed_messages
            }
        }

    @classmethod
    def from_data(cls, data: dict) -> AllowedMsgAllowance:
        allowance = data["allowance"]
        return cls(
            allowance=Allowance.from_data(allowance),
            allowed_messages=data["allowed_messages"],
        )

    def to_proto(self) -> AllowedMsgAllowance_pb:
        return AllowedMsgAllowance_pb(
            allowance=self.allowance.to_proto(), allowed_messages=self.allowed_messages
        )


#Allowance = Union[BasicAllowance, PeriodicAllowance]
class Allowance:#(BasicAllowance, PeriodicAllowance):

    @classmethod
    def from_data(cls, data: dict):
        if data.get("@type") == BasicAllowance.type_url:
            return BasicAllowance.from_data(data)
        else:
            return PeriodicAllowance.from_data(data)



import base64
import json
from typing import Any

from terra_sdk.core import Numeric

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncWasmAPI", "WasmAPI"]


class AsyncWasmAPI(BaseAsyncAPI):
    async def code_info(self, code_id: int) -> dict:
        """Fetches information about an uploaded code.

        Args:
            code_id (int): code ID

        Returns:
            dict: code information
        """
        res = await self._c._get(f"/terra/wasm/v1beta1/codes/{code_id}")
        code_info = res.get("code_info")
        return {
            "code_id": Numeric.parse(code_info["code_id"]),
            "code_hash": code_info["code_hash"],
            "creator": code_info["creator"],
        }

    async def contract_info(self, contract_address: str) -> dict:
        """Fetches information about an instantiated contract.

        Args:
            contract_address (str): contract address

        Returns:
            dict: contract information
        """
        res = await self._c._get(f"/terra/wasm/v1beta1/contracts/{contract_address}")
        contract_info = res.get("contract_info")
        return {
            "code_id": Numeric.parse(contract_info["code_id"]),
            "address": contract_info["address"],
            "creator": contract_info["creator"],
            "admin": contract_info.get("admin", None),
            "init_msg": contract_info["init_msg"],
        }

    async def contract_query(self, contract_address: str, query: dict) -> Any:
        """Runs a QueryMsg on a contract.

        Args:
            contract_address (str): contract address
            query_msg (dict): QueryMsg to run

        Returns:
            Any: results of query
        """
        params = {
            "query_msg": base64.b64encode(json.dumps(query).encode("utf-8")).decode(
                "utf-8"
            )
        }
        res = await self._c._get(
            f"/terra/wasm/v1beta1/contracts/{contract_address}/store", params
        )
        return res.get("query_result")

    async def parameters(self) -> dict:
        """Fetches the Wasm module parameters.

        Returns:
            dict: Wasm module parameters
        """
        res = await self._c._get("/terra/wasm/v1beta1/params")
        params = res.get("params")
        return {
            "max_contract_size": Numeric.parse(params["max_contract_size"]),
            "max_contract_gas": Numeric.parse(params["max_contract_gas"]),
            "max_contract_msg_size": Numeric.parse(params["max_contract_msg_size"]),
        }


class WasmAPI(AsyncWasmAPI):
    @sync_bind(AsyncWasmAPI.code_info)
    def code_info(self, code_id: int) -> dict:
        pass

    code_info.__doc__ = AsyncWasmAPI.code_info.__doc__

    @sync_bind(AsyncWasmAPI.contract_info)
    def contract_info(self, contract_address: str) -> dict:
        pass

    contract_info.__doc__ = AsyncWasmAPI.contract_info.__doc__

    @sync_bind(AsyncWasmAPI.contract_query)
    def contract_query(self, contract_address: str, query_msg: dict) -> Any:
        pass

    contract_query.__doc__ = AsyncWasmAPI.contract_query.__doc__

    @sync_bind(AsyncWasmAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncWasmAPI.parameters.__doc__



from terra_sdk.core.public_key import PublicKey
from terra_sdk.core.tx import TxInfo, TxLog, parse_tx_logs

from .account import Account
from .base_account import BaseAccount
from .lazy_graded_vesting_account import LazyGradedVestingAccount

__all__ = [
    "TxLog",
    "TxInfo",
    "PublicKey",
    "parse_tx_logs",
    "Account",
    "BaseAccount",
    "LazyGradedVestingAccount",
]



from __future__ import annotations

import base64
import hashlib
from abc import ABC, abstractmethod
from typing import List, Union

import attr

from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.crypto.ed25519 import PubKey as ValConsPubKey_pb
from terra_proto.cosmos.crypto.multisig import LegacyAminoPubKey as LegacyAminoPubKey_pb
from terra_proto.cosmos.crypto.secp256k1 import PubKey as SimplePubKey_pb

from .bech32 import get_bech
from terra_sdk.util.json import JSONSerializable

BECH32_AMINO_PUBKEY_DATA_PREFIX_SECP256K1 = "eb5ae987"+"21"  # with fixed length 21
BECH32_AMINO_PUBKEY_DATA_PREFIX_ED25519 = "1624de64"+"20"  # with fixed length 20
BECH32_AMINO_PUBKEY_DATA_PREFIX_MULTISIG_THRESHOLD = "22c1f7e2"  # without length


__all__ = ["PublicKey", "SimplePublicKey", "ValConsPubKey", "LegacyAminoMultisigPublicKey"]


def encode_uvarint(value: Union[int, str]) -> List[int]:
    val = int(str(value))
    if val > 127:
        raise ValueError('Encoding numbers > 127 is not supported here. Please tell those lazy CosmJS maintainers to port the binary.PutUvarint implementation from the Go standard library and write some tests.')
    return [val]


def address_from_public_key(public_key: PublicKey) -> bytes:
    sha = hashlib.sha256()
    rip = hashlib.new("ripemd160")
    sha.update(public_key.key)
    rip.update(sha.digest())
    return rip.digest()


def pubkey_from_public_key(public_key: PublicKey) -> bytes:
    arr = bytes.fromhex(BECH32_AMINO_PUBKEY_DATA_PREFIX_SECP256K1)
    arr += bytes(public_key.key)
    return bytes(arr)

class PublicKey(JSONSerializable, ABC):
    """Data object holding the public key component of an account or signature."""

    @abstractmethod
    def get_type(self) -> str:
        return self.type_url

    @classmethod
    def from_proto(cls, proto: Any_pb):
        type_url = proto.type_url
        if type_url == SimplePublicKey.type_url:
            return SimplePublicKey.from_proto(proto)
        elif type_url == ValConsPubKey.type_url:
            return ValConsPubKey.from_proto(proto)
        elif type_url == LegacyAminoMultisigPublicKey.type_url:
            return LegacyAminoMultisigPublicKey.from_proto(proto)
        raise TypeError(f"could not marshal PublicKey: type is incorrect")

    @classmethod
    def from_data(cls, data: dict):
        type_url = data["@type"]
        if type_url == SimplePublicKey.type_url:
            return SimplePublicKey.from_data(data)
        elif type_url == ValConsPubKey.type_url:
            return ValConsPubKey.from_data(data)
        elif type_url == LegacyAminoMultisigPublicKey.type_url:
            return LegacyAminoMultisigPublicKey.from_data(data)
        raise TypeError(f"could not unmarshal PublicKey: type is incorrect")

    @abstractmethod
    def pack_any(self) -> Any_pb:
        raise NotImplementedError

    @abstractmethod
    def address(self) -> str:
        pass

    @abstractmethod
    def raw_address(self) -> str:
        pass

    @abstractmethod
    def encode_amino_pubkey(self) -> bytes:
        pass


@attr.s
class SimplePublicKey(PublicKey):
    """Data object holding the SIMPLE public key component of an account or signature."""

    type_amino = "tendermint/PubKeySecp256k1"
    """"""

    type_url = "/cosmos.crypto.secp256k1.PubKey"
    """Normal signature public key type."""

    key: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": self.key
        }

    def to_data(self) -> dict:
        return {"key": self.key}

    @classmethod
    def from_data(cls, data: dict) -> SimplePublicKey:
        return cls(key=data["key"])

    def to_proto(self) -> SimplePubKey_pb:
        return SimplePubKey_pb(key=self.key)

    def get_type(self) -> str:
        return self.type_url

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    def encode_amino_pubkey(self) -> bytearray:
        out = bytearray.fromhex(BECH32_AMINO_PUBKEY_DATA_PREFIX_SECP256K1) + bytearray(self.key)
        return out

    def raw_address(self) -> str:
        return address_from_public_key(self.key)

    def address(self) -> str:
        return get_bech("terra", self.raw_address())


@attr.s
class ValConsPubKey(PublicKey):
    """Data object holding the public key component of an validator's account or signature."""

    type_amino = "tendermint/PubKeyEd25519"
    """"""

    type_url = "/cosmos.crypto.ed25519.PubKey"
    """an ed25519 tendermint public key type."""

    key: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": self.key
        }

    def to_data(self) -> dict:
        return {"key": self.key}

    @classmethod
    def from_data(cls, data: dict) -> ValConsPubKey:
        return cls(key=data["key"])

    def get_type(self) -> str:
        return self.type_url

    def to_proto(self) -> ValConsPubKey_pb:
        return ValConsPubKey_pb(key=base64.b64encode(self.key))

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    def encode_amino_pubkey(self) -> bytes:
        return bytes.fromhex(BECH32_AMINO_PUBKEY_DATA_PREFIX_ED25519) + bytes(self.key)

    def raw_address(self) -> str:
        return address_from_public_key(self.key)

    def address(self) -> str:
        return get_bech("terravalcons", self.raw_address())


@attr.s
class LegacyAminoMultisigPublicKey(PublicKey):
    """Data object holding the Legacy Amino-typed public key component of an account or signature."""

    type_amino = "tendermint/PubKeyMultisigThreshold"
    """"""

    type_url = "/cosmos.crypto.multisig.LegacyAminoPubKey"
    """Multisig public key type."""

    threshold: int = attr.ib(converter=int)
    public_keys: List[SimplePublicKey] = attr.ib(factory=list)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "threshold": str(self.threshold),
                "pubkeys": [pubkey.to_amino() for pubkey in self.public_keys]
            }
        }

    def to_data(self) -> dict:
        return {"threshold": self.threshold, "public_keys": self.public_keys}

    @classmethod
    def from_data(cls, data: dict) -> LegacyAminoMultisigPublicKey:
        return cls(threshold=data["threshold"], public_keys=data["public_keys"])

    def get_type(self) -> str:
        return self.type_url

    def to_proto(self) -> LegacyAminoPubKey_pb:
        return LegacyAminoPubKey_pb(
            threshold=self.threshold,
            public_keys=[pk.pack_any() for pk in self.public_keys]
        )

    def encode_amino_pubkey(self) -> bytearray:
        if self.threshold > 127 :
            raise ValueError("threshold over 127 is now supported here")
        out = bytearray.fromhex(BECH32_AMINO_PUBKEY_DATA_PREFIX_MULTISIG_THRESHOLD)
        out.append(0x08)
        out += bytearray(encode_uvarint(self.threshold))
        for pkData in [pubkey.encode_amino_pubkey() for pubkey in self.public_keys]:
            out.append(0x12)
            out += bytearray(encode_uvarint(len(pkData)))
            out += pkData
        return out

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    def raw_address(self) -> str:
        pubkey_data = bytes(self.encode_amino_pubkey())
        hasher = hashlib.sha256()
        hasher.update(pubkey_data)
        return hasher.digest()[0:20].hex()

    def address(self) -> str:
        address = get_bech("terra", self.raw_address())
        return address

    def pubkey_address(self) -> str:
        #return bech32.bech32_encode('terrapub', bech32.convertbits(bytes(self.raw_address()), 8, 5))
        return get_bech("terrapub", str(self.encode_amino_pubkey()))



"""Oracle module data objects."""

from __future__ import annotations

import attr
from terra_proto.terra.oracle.v1beta1 import (
    AggregateExchangeRatePrevote as AggregateExchangeRatePrevote_pb,
)
from terra_proto.terra.oracle.v1beta1 import (
    AggregateExchangeRateVote as AggregateExchangeRateVote_pb,
)

from terra_sdk.core import Coin, Coins, ValAddress
from terra_sdk.util.json import JSONSerializable

__all__ = [
    "AggregateExchangeRatePrevote",
    "AggregateExchangeRateVote",
]


@attr.s
class AggregateExchangeRateVote(JSONSerializable):
    """Contains information about a validator's aggregate vote."""

    exchange_rate_tuples: Coins = attr.ib(converter=Coins)
    """Reported exchange rates by validator."""

    voter: ValAddress = attr.ib()
    """Validator that sent the aggregate vote."""

    def to_amino(self) -> dict:
        tuples = self.exchange_rate_tuples.to_amino()
        return {
            "exchange_rate_tuples": [{"denom": x.denom, "exchange_rate":str(x.amount)} for x in tuples],
            "voter": self.voter
        }

    def to_data(self) -> dict:
        return {
            "exchange_rate_tuples": [
                {"denom": x.denom, "exchange_rate": str(x.amount)}
                for x in self.exchange_rate_tuples
            ],
            "voter": self.voter,
        }

    @classmethod
    def from_data(cls, data) -> AggregateExchangeRateVote:
        return cls(
            exchange_rate_tuples=Coins(
                [
                    Coin(d["denom"], d["exchange_rate"])
                    for d in data["exchange_rate_tuples"]
                ],
            ),
            voter=data["voter"],
        )

    def to_proto(self) -> AggregateExchangeRateVote_pb:
        return AggregateExchangeRateVote_pb(
            exchange_rate_tuples=[
                {"denom": tuple.denom, "exchange_rate": str(tuple.amount)}
                for tuple in self.exchange_rate_tuples.to_list()
            ],
            voter=self.voter,
        )


@attr.s
class AggregateExchangeRatePrevote(JSONSerializable):
    """Contains information about a validator's aggregate prevote."""

    hash: str = attr.ib()
    """Aggregate vote hash for the upcoming aggregate vote."""

    voter: ValAddress = attr.ib()
    """Validator that submitted the aggregate prevote."""

    submit_block: int = attr.ib(converter=int)
    """Block height at which the aggregate prevote was submitted."""

    def to_amino(self) -> dict:
        return {
            "hash": self.hash,
            "voter": self.voter,
            "submit_block": str(self.submit_block)
        }

    def to_data(self) -> dict:
        return {
            "hash": self.hash,
            "voter": self.voter,
            "submit_block": str(self.submit_block)
        }

    @classmethod
    def from_data(cls, data) -> AggregateExchangeRatePrevote:
        return cls(
            hash=data["hash"],
            voter=data["voter"],
            submit_block=int(data["submit_block"]),
        )

    def to_proto(self) -> AggregateExchangeRatePrevote_pb:
        return AggregateExchangeRatePrevote_pb(
            hash=self.hash, voter=self.voter, submit_block=self.submit_block
        )
